
"""levenshtein.py - a module for computing minimum string edit distance.

Brett Kessler, Washington University Linguistics

>>> from levenshtein import Levenshtein
>>> lev = Levenshtein()
>>> lev.set_pair('intention', 'execution')
>>> computed_distance = lev.min_edit_distance()
>>> print lev.alignment()
i >  
n ~ e
t ~ x
e = e
  < c
n ~ u
t = t
i = i
o = o
n = n
>>> lev.table(sys.stdout)
   #   i   n   t   e   n   t   i   o   n 
#  00 →01 →02 →03 →04 →05 →06 →07 →08 →09
    ↓ ↘   ↘   ↘   ↘                      
e  01  02  03  04  03 →04 →05 →06 →07 →08
    ↓ ↘   ↘   ↘    ↓  ↘   ↘   ↘   ↘   ↘  
x  02  03  04  05  04  05  06  07  08  09
    ↓ ↘   ↘   ↘   ↘   ↘   ↘   ↘   ↘   ↘  
e  03  04  05  06  05  06  07  08  09  10
    ↓ ↘   ↘   ↘    ↓  ↘   ↘   ↘   ↘   ↘  
c  04  05  06  07  06  07  08  09  10  11
    ↓ ↘   ↘   ↘    ↓  ↘   ↘   ↘   ↘   ↘  
u  05  06  07  08  07  08  09  10  11  12
    ↓ ↘   ↘   ↘       ↘   ↘              
t  06  07  08  07 →08  09  08 →09 →10 →11
    ↓ ↘           ↘   ↘    ↓  ↘          
i  07  06 →07 →08  09  10  09  08 →09 →10
    ↓  ↓  ↘   ↘   ↘   ↘    ↓   ↓  ↘      
o  08  07  08  09  10  11  10  09  08 →09
    ↓  ↓  ↘           ↘        ↓   ↓  ↘  
n  09  08  07 →08 →09  10 →11  10  09  08

Illustration is from Jurafsky & Martin, 2nd ed., p. 73–77.

Default engine is set up to give a cost of 1 to all insertions and deletions,
2 to all nonidentical substitutions. To change these values, inherit the
Levenshtein class and override these methods, which return a numeric
value for the cost:
       
ins_cost(self, element)
del_cost(self, element)
sub_cost(self, source, target)

The last should return 0 for identical elements, but that can be changed if
you are not concerned that what the algorithm finds are technically
distances.

"""

import unicodedata

version = '0.2'

class LevenshteinError(Exception):
    """An error encountered by this program."""
    pass

START = 'start'
INSERT = 'i'
DELETE = 'd'
SUBSTITUTE = 's'

def print_len(string):
    """Return length of string in characters, excluding Mark characters
        (diacritics).
        This is meant to approximate width of text in
        a monospaced fount, so that a plaintext table has a decent
        chance of lining up the columns correctly.
    """
    width = 0
    for character in string:
        if not unicodedata.category(character).startswith('M'):
            width += 1
    return width
    
    
class Cell(object):
    """A cell in a distance matrix. Each element tells the cumulative cost
        of getting to this point in the table (.num) and what operation
        was taken to get here – SUBSTITUTION, DELETION, or INSERTION.
        The value START is used in the upper left cell (0, 0)."""

    def __init__(self, num, op):
        self.num = num
        self.op = op
    
    
class Levenshtein(object):
    """An engine for computing string-edit distance. The object keeps
            state for each comparison until a new pair of strings is set.
        self.set_pair(source, target) - which strings to compare.
        self.min_edit_distance() - value for last-set pair of strings
        self.table(stream) - writes to Output stream a plaintext
            table showing numbers and best operation for each cell.
        self.alignment() - returns string showing optimal alignment
        self.alignment_facts() - returns list of tuples showing alignment
            in more programmatic form
        
        Private properties:
        self.source - list of objects to be treated as source
        self.target - ... and as target
        self.widest_source_char, .widest_target_char - plaintext print widths
            of the biggest source character and target character
        self.distance - dict maps [target_pos][source_pos] to a Cell
        self.compute_min_edit_distance() - does the work
    """
    
    def __init__(self):
        self.source = []
        self.target = []
        self.distance = None
        
    @staticmethod
    def widest(string):
        """Find the printing length of the longest element in the list."""
        widest = 0
        for element in string:
            widest = max(widest, print_len(element))
        return widest
    
    def set_pair(self, source, target):
        """Set the two items to be prepared. Arguments can be lists of
            arbitrary objects (a shallow copy is made), or strings."""
        self.source = list(source)
        self.target = list(target)
        self.widest_source_char = Levenshtein.widest(self.source)
        self.widest_target_char = Levenshtein.widest(self.target)
        self.compute_min_edit_distance()
        
    def ins_cost(self, element):
        """Return the cost of inserting this element."""
        return 1
        
    def del_cost(self, element):
        """Return the cost of deleting this element."""
        return 1
        
    def sub_cost(self, source, target):
        """Return the cost of replacing source element with target element.
            By definition, distance ought to be 0 if the two elements are
            the same.
        """
        return 0 if source == target else 2
        
    def compute_min_edit_distance(self):
        """Compute the distance between source and target."""
        target = self.target
        source = self.source
        
        # Initialize margins of distance matrix:
        self.distance = distance = {(0, 0): Cell(0, START)}
        for targ_pos, targ_element in enumerate(target):
            i = targ_pos + 1
            distance[i, 0] = Cell(
                distance[targ_pos, 0].num + self.ins_cost(targ_element),
                INSERT)
        for src_pos, src_element in enumerate(source):
            j = src_pos + 1
            distance[0, j] = Cell(
                distance[0, src_pos].num + self.del_cost(src_element),
                DELETE)
                
        # Compute body of distance matrix:
        for targ_pos, targ_element in enumerate(target):
            i = targ_pos + 1
            for src_pos, src_element in enumerate(source):
                j = src_pos + 1
                ins_cost = (
                    distance[targ_pos, j].num +
                    self.ins_cost(targ_element))
                sub_cost = (
                    distance[targ_pos, src_pos].num +
                    self.sub_cost(src_element, targ_element))
                del_cost = (
                    distance[i, src_pos].num +
                    self.del_cost(src_element))
                min_cost = min(ins_cost, sub_cost, del_cost)
                distance[i, j] = Cell(min_cost,
                     SUBSTITUTE if min_cost == sub_cost else
                     DELETE if min_cost == del_cost else
                     INSERT)
                    
    def min_edit_distance(self):
        """Return the distance between source and target."""
        return self.distance[len(self.target), len(self.source)].num
        
    def table(self, stream):
        """Make plaintext table showing the calculations"""
        source = self.source
        target = self.target
        distance = self.distance
        # Compute length of widest entry:
        widest = max(self.widest_source_char, self.widest_target_char)
        for cell in distance.values():
            widest = max(widest, print_len(cell.num))
        ws = '%%0%ii' % widest
        # Top labels (source string):
        stream.write(' ' * widest)  # column for target elements
        stream.write(' ')
        stream.write('#'.center(widest))
        for element in source:
            stream.write('  ')
            stream.write(unicode(element).center(widest))
        stream.write('\n')
        # 2nd line (source margin):
        stream.write('#'.center(widest))
        stream.write(' ')
        stream.write(ws % 0)
        for pos in range(1, len(source) + 1):
            stream.write(' →')
            cell = distance[0, pos]
            stream.write((ws % cell.num).center(widest))
        stream.write('\n')
        # Subsequent lines:
        for targ_pos, targ_element in enumerate(target):
            i = targ_pos + 1
            # Top line just for ↓ and ↘ :
            stream.write(' '.center(widest) * 2)
            stream.write('↓')
            for j in range(1, len(source) + 1):
                cell = distance[i, j]
                stream.write(' ')
                stream.write('↘' if cell.op == SUBSTITUTE else ' ')
                stream.write(('↓' if cell.op == INSERT else ' ').center(
                    widest))
            stream.write('\n')
            # Second line contains character, →, numbers:
            stream.write(targ_element.center(widest))
            stream.write(' ')
            stream.write((ws % distance[i, 0].num).center(widest))
            for j in range(1, len(source) + 1):
                cell = distance[i, j]
                stream.write(' ')
                stream.write('→' if cell.op == DELETE else ' ')
                stream.write((ws % cell.num).center(
                    widest))
            stream.write('\n')
            
    def alignment_facts(self):
        """Return a list of tuples showing the optimal alignment.
            Each tuple is (source_char, target_char, cost),
            where source_char can be None (insertion) and target_char
            can be None (deletion). cost tells cumulative cost.
           Works by backtracing from the last cell – where input is
            totally consumed and output totally generated – back to
            the origin.
        """
        backtrace = []
        source = self.source
        target = self.target
        distance = self.distance
        s = len(source)
        t = len(target)
        while s > 0 or t > 0:
            cell = distance[t, s]
            provenance = cell.op
            cost = cell.num
            if provenance == SUBSTITUTE:
                s -= 1
                t -= 1
                s_char = source[s]
                t_char = target[t]
                backtrace.insert(0, (s_char, t_char, cost))
            elif provenance == INSERT:
                t -= 1
                t_char = target[t]
                backtrace.insert(0, (None, t_char, cost))
            else:  # provenance == DELETE:
                s -= 1
                s_char = source[s]
                backtrace.insert(0, (s_char, None, cost))
        return backtrace
            
    def alignment(self):
        """Return the alignment as a printable plaintext string. Each 
            alignment is separated from the following by a newline.
            It shows source character (white space if insertion),
            operation (< for insertion, > for deletion, ~ for substitution,
            = for no change), target character (white space if deletion),
            and cost of the operation."""
        bt = self.alignment_facts()
        lines = []
        format = '%%%is %%s %%%is  %%s' % (
            self.widest_source_char, self.widest_target_char)
        prev_cost = 0
        for source, target, cum_cost in bt:
            cost = cum_cost - prev_cost
            lines.append(format % (
                (source or ' '),
                ('>' if target is None else
                 '<' if source is None else
                 '=' if source == target else
                 '~'),
                (target or ' '),
                cost))
            prev_cost = cum_cost
        return '\n'.join(lines)

