#! /usr/bin/env python3
# encoding: UTF-8

"""ponto.cgi.py

Run as CGI to service Ponto requests.

"""

from cgi import escape as html_escape
from datetime import datetime
from os import environ
import re
from sys import stderr, stdin, stdout
# These are TextIOWrap'ed with ANSI_X3.4-1968.
import unicodedata

from ponto import Ponto

class PontoError(Exception): pass

def set_up_output():
    """We will use the byte I/O layer of stdout and explicitly encode
        as needed. Probably makes most sense to just insist on UTF-8. 
        But if we find a need to intelligently handle other encodings, and a
        reasonable way to do so, we'd set it here. (NB there is no other 
        well-known encoding for IPA, so it would have to be something with no
        pronunciations, or SAMPA, etc.)"""
    global encoding_out
    encoding_out = 'UTF-8'
    global out_stream
    out_stream = stdout.buffer

def write_http_header():
    out_stream.write(b'Content-Type: application/xhtml+xml; charset=')
    out_stream.write(encoding_out.encode('ascii'))
    out_stream.write(b'\n\n')

def output(string):
    """Writes a Unicode str to stdout (i.e., back to the client),
        in the appropriate encoding as set in set_up_output()"""
    out_stream.write(string.encode(encoding_out, 'xmlcharrefreplace'))
    
def write_html_header(title):
    output('''<!DOCTYPE html 
    PUBLIC "-//W3C//DTD XHTML 1.1//EN"
    "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" dir="ltr">
<head>
<meta http-equiv="Content-Type" 
    content="application/xhtml+xml; charset={charset}"/>
<title>{title}</title>
<style type="text/css" media="screen">
@import url(../ponto.css);
</style>
</head>
<body>
<p id="banner">
<span id="WUStL"><a
    href="http://www.wustl.edu/">Washington University in St. Louis</a>
</span>
/
<span><a href="http://www.artsci.wustl.edu/~psych/">Psychology</a></span>
/
<span><a href="http://spell.psychology.wustl.edu">Brett Kessler</a></span>
</p>

<h1>{title}</h1>
'''.format(title=title, charset=encoding_out))

def read_form_bytes():
    """Read the submitted data, load into a table called "form". 
        form[str] = b'bytes'.  We don’t do decoding data to Unicode now, 
        in case later we implement some process of choosing alternative
        encodings based on information in the form itself.
        Data format of form is POST multipart/form-data because this can
        include huge files."""
    global form; form = {}  # name (Unicode) -> value (bytes)
    in_stream = stdin.buffer
    # To do: consider setting an upper limit on how many bytes we'll take:
    multipart = b'\r\n' + in_stream.read()
    if len(multipart) <= 2:
        raise PontoError('No form data encountered (no stdin). ' \
            'Please submit via the ponto.html search page.')
    content_type = environ.get('CONTENT_TYPE', None)
    if content_type is None:
        raise PontoError('No form data encountered (no CONTENT_TYPE). ' \
            'Please submit via the ponto.html search page.')
    match = re.match(
        b'multipart/form-data; \s boundary=(?P<boundary> .+) \Z',
        content_type.encode('ANSI_X3.4-1968'),
        re.VERBOSE)
    if match is None:
        raise PontoError('No multipart/form-data. ' \
            'Please submit via the ponto.html search page.')
    boundary = b'\r\n--' + match.group('boundary')
    sections = multipart.split(boundary)[1:-1]
    if len(sections) < 1:
        raise PontoError('No form-data items. ' \
            'Please submit via the ponto.html search page.')
    header_re = re.compile(
        b'\r\nContent-Disposition: form-data; name="(?P<name>[^"]+)"\r\n\r\n')
    for section in sections:
        match = header_re.match(section)
        if match is None:
            raise PontoError('Unrecognized form-data item:\n' + 
                section[0:100].decode('ASCII'))
        name = match.group('name').decode('UTF-8')
        form[name] = section[match.end():]

def decode_form_data(field_name):
    """Reads the named entry from the form, or empty string if not in
        form. We always decode to Unicode NFD and strip marginal spaces."""
    if field_name not in form: return ''
    # For now, let it raise error if input is not UTF-8 (or ASCII);
    # we’ll catch it later and exit:
    return unicodedata.normalize('NFD',
        form.get(field_name, '').decode(encoding_in).strip())

def read_string(name):
    """Reads the named entry from the form, or raises an error message
        if not found. Decode to Unicode and strip marginal spaces."""
    if name not in form:
        raise PontoError('Missing form-data item "{0}".'
            ' Please submit via the ponto.html search page.'.format(name))
    return decode_form_data(name)

def read_multiple_select(name):
    """Read the select input, returning a list of values"""
    if name not in form: return []
    value_str = read_string(name)
    return value_str.split(',')

def read_checkbox(name):
    """Reads and returns True if checkbox checked in form. There will be a 
        form field with the given name and value 'on'. If not checked, there
        will be no form field transmitted at all. If it appears with any
        value besides 'on', there's a problem."""
    if name not in form: return False
    value = form[name]
    if value == b'on': return True
    raise PontoError('Checkbox {0} transmitted with unknown value.'
        ' Please submit via the ponto.html search page.'.format(name))

def read_float(name, default):
    """Reads named value from form and converts it to a float. If can't be
        interpreted as a float, raises an error. If empty, returns default
        value"""
    float_str = read_string(name)
    if not float_str: return default
    try:
        value = float(float_str)
    except ValueError as msg:
        raise PontoError(
            'Value of {0} field should be a number, not {0}'.format(float_str))
    return value
            
def guess_column_separator(string):
    """Guess what separator user used to separate cells in a row of data"""
    return '\t' if '\t' in string \
        else ',' if ',' in string \
        else ' ' if ' ' in string \
        else '\t'
        
def cell_strip(content):
    """Strips a table cell of marginal space and surrounding quotes."""
    content = content.strip()
    if (content.startswith('"') and content.endswith('"')) or \
       (content.startswith("'") and content.endswith("'")):
        content = content[1:-1].strip()
    return content

def read_table(name, required_columns):
    """Reads a table that was entered as a textarea. We guess what the
        delimiters are that separate rows and columns by picking the first
        that is present: CR/LF, CR, or LF to separate rows;
        HT, comma, or space to separate columns. Marginal spaces are
        deleted from each cell. The first row is treated as column heads.
        Returns dict with 'column_heads' (list of str) and 'rows' (list of
        list of str)"""
    table = read_string(name)
    if not table or table.isspace(): return {'rows': []}
    eol = '\r\n' if '\r\n' in table \
        else '\r' if '\r' in table \
        else '\n'
    lines = table.split(eol)
    header = lines.pop(0)
    field_separator = guess_column_separator(header)
    column_heads = [cell_strip(name) for name in header.split(field_separator)]
    for col in required_columns:
        if col not in column_heads:
            raise PontoError('{table} table needs a {col} column'.format(
                table=name, col=col))
    n_columns = len(column_heads)
    rows = []
    for line in lines:
        if not line: continue
        if line.isspace(): continue
        field_separator = guess_column_separator(line)
        cells = line.strip().split(field_separator)
        n_cells = len(cells)
        if n_cells < n_columns:
            cells.extend([''] * (n_columns - n_cells))
        row = [cell_strip(cell) for cell in cells]
        rows.append(row)
    return {'eol': eol, 'column_heads': column_heads, 'rows': rows}

def read_trials():
    """Reads the Trials textarea as a table (see read_table()), raising an
        error if omitted."""
    trials = read_table('trials', ('stimulus', 'response'))
    if not trials['rows']:
        raise PontoError('Enter your data into Trials in the form of a table, one'
            ' row per trial. First row must be column names. Separate columns'
            ' with a Tab, comma, or space. Table must have columns'
            ' “stimulus” (word subject is trying to write; separate'
            ' alternative interpretations of a stimulus with |) and'
            ' “response” (subject’s attempt; use --- for missing trials, ""'
            ' for zero-length productions), and, if by-subject summaries are'
            ' wanted, “subject” (ID of participant). Other columns are'
            ' ignored.')
    return trials

def write_trials_parameters(trials):
    """Writes a summary of trials information"""
    output('<h3>Trials</h3>')
    output('<p>Trials contain {0} rows</p>'.format(len(trials['rows'])))

def read_correspondences_from_form(corrs):
    """Reads the Correspondences Textarea of the form and translates it into
        a dict that maps  stimulus -> response -> penalty"""
    table = read_table('correspondences', ('stimulus', 'response', 'penalty'))
    if 'column_heads' not in table: return
    stimulus_pos = table['column_heads'].index('stimulus')
    response_pos = table['column_heads'].index('response')
    penalty_pos = table['column_heads'].index('penalty')
    required_columns = max(stimulus_pos, response_pos, penalty_pos)
    cmap = corrs['map']
    for row in table['rows']:
        if len(row) <= required_columns:
            raise PontoError(
                'Correspondences table has too-short row: {0}'.format(row))
        stimulus = row[stimulus_pos]
        if stimulus not in cmap:
            cmap[stimulus] = {}
        stimulus_table = cmap[stimulus]
        response = row[response_pos]
        penalty_str = row[penalty_pos]
        try:
            penalty = float(penalty_str)
        except ValueError:
            raise PontoError(
                'Correspondences penalty must be a number, not {0}'.format(
                    penalty))
        stimulus_table[response] = penalty

corr_re = re.compile(
    '<tr><td>(?P<stim>.*?)</td><td>(?P<resp>.*?)</td><td>(?P<pen>.*?)</td>')
comment_re = re.compile('<!--.*?-->')

def add_correspondences_boilerplate(corrs, boiler_name, boiler_file):
    """Adds to correspondences those defined in a specific boilerplate,
        unless already specified.
        To do: use dom reader."""
    with open('../' + boiler_file, encoding='UTF-8') as in_file:
        state = 'before'
        cmap = corrs['map']
        for line in in_file.readlines():
            line = comment_re.sub('', line)
            if line.isspace(): continue
            line = line.rstrip()
            if state == 'before' and '<table class="correspondences">' in line:
                state = 'inside'
                continue
            if state == 'inside':
                if '</table>' in line: break
                if '<tr><th>' in line: continue
                match = corr_re.match(line)
                if not match: 
                    raise PontoError('Bad line in {file} table: {line}'
                        .format(file=boiler_file, line=line))
                stimulus = unicodedata.normalize('NFD', match.group('stim'))
                if stimulus not in cmap:
                    cmap[stimulus] = {}
                stimulus_table = cmap[stimulus]
                response = unicodedata.normalize('NFD', match.group('resp'))
                if response in stimulus_table: continue
                penalty = float(match.group('pen'))
                stimulus_table[response] = penalty
    corrs['boilerplates'].add(boiler_name)
            
def add_correspondences_boilerplates(corrs):
    """Adds to correspondences the ones defined by boilerplates"""
    boiler_names = read_multiple_select('correspondence_files')
    if 'identity' in boiler_names:
        corrs['accept-identity'] = True
        boiler_names.remove('identity')
    for boiler_name, boiler_file in (
            ('ampr-ipa', 'ampr-ipa.html'),
            ('ampr-sampa', 'ampr-sampa.html'),
            ('ampr-eng', 'ampr-eng.html'),
            ('gr1syl1', 'gr1syl1.html'),
            ('heather', 'heather-ipa.html'),
            ('tatiana', 'tatiana-ipa.html'),
            ('ces-nonex', 'ces-nonex.html'),
            ('eng-rp-nonex', 'eng-rp-nonex.html'),
            ('spa-nonex', 'spa-nonex.html'),
            ('fra-nonex', 'fra-nonex.html'),
            ('fra-manulex', 'fra-manulex.html'),
            ('slk-nonex', 'slk-nonex.html')
            ):
        if boiler_name in boiler_names:
            add_correspondences_boilerplate(corrs, boiler_name, boiler_file)

def read_default_penalties(corrs):
    """Loads default penalties values from form"""
    corrs['insertion-penalty'] = read_float('insertion-penalty', 1.0)
    corrs['deletion-penalty'] = read_float('deletion-penalty', 1.0)
    corrs['substitution-penalty'] = read_float('substitution-penalty', 1.0)

def read_correspondences():
    """Reads the Correspondences section of the form"""
    correspondences = {
        'map': {}, 'accept-identity': None, 'boilerplates': set()}
    read_correspondences_from_form(correspondences)
    add_correspondences_boilerplates(correspondences)
    read_default_penalties(correspondences)
    return correspondences

def write_correspondences_parameters(corrs):
    """Echoes back summary of correspondences settings"""
    output('''<h3>Correspondences</h3>
    <p>All correspondences:</p>
    <table>
    <tr><th>stimulus</th><th>response</th><th>penalty</th></tr>
''')
    cmap = corrs['map']
    for stim in sorted(cmap):
        stim_table = cmap[stim]
        for resp in sorted(stim_table):
            penalty = stim_table[resp]
            output('<tr><td>{stim}</td><td>{resp}</td><td>{pen}</td></tr>\n'
                .format(
                    stim=html_escape(stim),
                    resp=html_escape(resp),
                    pen=penalty))
    output('</table>')
    if corrs['accept-identity']:
        output('<p>Plus all identity matches not listed above.</p>')
    if corrs['boilerplates']:
        output('<p>Above list incorporates boilerplate correspondences in'
            ' {0}</p>'.format(' and '.join(corrs['boilerplates'])))
    output('<p>Default scores for other correspondences:</p>\n<ul>\n')
    for default in ('insertion-penalty', 
            'deletion-penalty', 
            'substitution-penalty'):
        output('<li>{name}: {pen}</li>\n'.format(
            name=default, pen=corrs[default]))
    output('</ul>\n')

def read_scoring():
    """Reads the Scoring section of the form"""
    return {
        'require-order': read_checkbox('require-order'),
        'do-monte': read_checkbox('do-monte')} 

def write_scoring_parameters(scoring):
    """Echoes the Scoring section of the form"""
    output('<h3>Scoring</h3>\n')        
    output('<p>Responses {0}have to be in the correct sequence.</p>\n'
        .format('' if scoring['require-order'] else 'do not '))       
    output('<p>{0} Monte Carlo significance test.</p>\n'
        .format('Run' if scoring['do-monte'] else 'Do not run'))

def read_form():
    """Reads the form into a dict giving all the data, as Unicode strings."""
    read_form_bytes()
    # What is the real encoding transmitted by the browser? I'm hoping it is
    # really UTF-8, which we asked for.
    global encoding_in; encoding_in = 'UTF-8'
    trials = read_trials()
    correspondences = read_correspondences()
    scoring = read_scoring()
    return {
        'trials': trials,
        'correspondences': correspondences,
        'scoring': scoring}
        
def write_html_footer():
    output('</body></html>\n')

def write_parameters(parameters):
    output('<h2>Parameters</h2>')
    output('<p>{0}</p>'.format(datetime.utcnow().isoformat()))
    write_trials_parameters(parameters['trials'])
    write_correspondences_parameters(parameters['correspondences'])
    write_scoring_parameters(parameters['scoring'])

def print_list_as_tr(list_):
    output('<tr>')
    for cell in list_:
        css_class = None
        if isinstance(cell, float):
            display = '{0:.3f}'.format(cell)
            css_class = 'num'
        elif isinstance(cell, int):
            display = str(cell)
            css_class = 'num'
        else:
            display = html_escape(str(cell))
            css_class = 'str'
        output('<td class="{1}">{0}</td>'.format(display, css_class))
    output('</tr>\n')
    out_stream.flush()

def print_list_as_header(list_, tips):
    output('<tr>')
    for header in list_:
        output('<td{0}>{1}</td>'.format(
            ' title="{0}"'.format(html_escape(tips[header])) if header in tips
                else '',    
            html_escape(header)))
    output('</tr>\n')
    
def do_scoring_table(parameters):
    """Call Ponto to do the work."""
    ponto = Ponto(
        trials=parameters['trials'],
        correspondences=parameters['correspondences'],
        scoring=parameters['scoring'])
    output('<h2>Scores</h2>\n')
    try:
        output('<table>\n')
        header = ponto.column_heads()
        print_list_as_header(header, {
            'distance': 'Penalties summed across all errors in the spelling;'
                            ' 0 means fully correct',
            'align': 
                'Way of matching up stimulus units and response units that'
                ' gives the best score. Each match is of format'
                ' [stimulus=response]. Asterisks * represent an error.'})
        for row in ponto.rows():
            print_list_as_tr(row)
    finally:
        output('</table>\n')
    output('<h2>Summaries</h2>')
    try:
        output('<table>\n')
        header = ponto.summary_column_heads()
        print_list_as_header(header, {
            'distSum': 'Sum of distance measures (errors) across all stimuli',
            'nTrials': 'Number of trials: how many words the subject spelled',
            'distMean': 'Average distance (error measure) per trial',
            'distRand':
                'Sum of distance measures (errors) expected if responses'
                ' were randomly assigned to stimuli',
            'improve':
                'Proportional improvement in attested distSum over chance'
                ' distRand'})
        for row in ponto.summary_rows():
            print_list_as_tr(row)
    finally:
        output('</table>\n')

def print_row(row):
    for num, cell in enumerate(row):
        if num > 0: output('\t')
        if isinstance(cell, float):
            cell = '{0:.3f}'.format(cell)
        else:
            cell = html_escape(str(cell))
        output(cell)
    output('\n')
    out_stream.flush()
   
def do_scoring(parameters):
    """Call Ponto to do the work. Return table inside <PRE>"""
    ponto = Ponto(
        trials=parameters['trials'],
        correspondences=parameters['correspondences'],
        scoring=parameters['scoring'])
    output('<h2>Scores</h2>\n')
    try:
        output('<pre>\n')
        header = ponto.column_heads()
        output('\t'.join(header))
        output('\n')
        for row in ponto.rows():
            print_row(row)
    finally:
        output('</pre>\n')
    output('<h2>Summaries</h2>')
    try:
        output('<pre>\n')
        header = ponto.summary_column_heads()
        output('\t'.join(header))
        output('\n')
        for row in ponto.summary_rows():
            print_row(row)
    finally:
        output('</pre>\n')
        
def main():
    problem = None
    #parameters = read_form()
    try:
        parameters = read_form()
    except Exception as msg:
        problem = str(msg)
    set_up_output()
    write_http_header()
    write_html_header('Error' if problem else 'Ponto Reply')
    if problem:
        output('<p>{0}</p></body></html>\n'.format(
            html_escape(problem)))
        raise SystemExit
    try:
        write_parameters(parameters)
        do_scoring(parameters)
    except Exception as msg:
        output('<h1>Error</h1>\n<p>{0}</p>\n'.format(html_escape(str(msg))))
    finally:
        write_html_footer()
    
main()

