#!/usr/local/bin/python3.0
# encoding: UTF-8

"""ponto.py

Scoring methods for tables of spelling data.
These are context-free.

"""

import random

from levenshtein import Levenshtein

class PontoError(Exception): pass

class PontoLevenshtein(Levenshtein):

    def __init__(self, 
            penalties, 
            insertion_penalty, deletion_penalty, substitution_penalty,
            accept_identity):
        super().__init__()
        self.penalties = penalties
        self.insertion_penalty = insertion_penalty
        self.deletion_penalty = deletion_penalty
        self.substitution_penalty = substitution_penalty
        self.accept_identity = accept_identity
    
    def get_penalty(self, stimulus, response):
        """Return the cost of aligning stimulus unit with response unit"""
        penalties = self.penalties
        if stimulus not in penalties: return None
        return penalties[stimulus].get(response, None)
    
    def ins_cost(self, element):
        """Return the cost of inserting this element."""
        penalty = self.get_penalty('', element)
        if penalty is None: penalty = self.insertion_penalty
        return penalty
        
    def del_cost(self, element):
        """Return the cost of deleting this element."""
        penalty = self.get_penalty(element, '')
        if penalty is None: penalty = self.deletion_penalty
        return penalty
        
    def sub_cost(self, stimulus, response):
        """Return the cost of replacing source element with target element.
            By definition, distance ought to be 0 if the two elements are
            the same.
        """
        penalty = self.get_penalty(stimulus, response)
        if penalty is None:
            if self.accept_identity and stimulus == response: penalty = 0.0
            else: penalty = self.substitution_penalty
        return penalty
        

class Ponto:
    
    def __init__(self, trials, correspondences, scoring):
        self.trials = trials
        self.correspondences = correspondences
        self.stimulus_units = self.correspondences['map']
        self.hash_response_units()
        self.scoring = scoring
        self.subjects = {}
        if self.scoring['require-order']:
            self.levenshtein = PontoLevenshtein(
                penalties=self.correspondences['map'],
                insertion_penalty=self.correspondences['insertion-penalty'],
                deletion_penalty=self.correspondences['deletion-penalty'],
                substitution_penalty=
                    self.correspondences['substitution-penalty'],
                accept_identity=self.correspondences['accept-identity'])
    
    def hash_response_units(self):
        """Make self.response_units a set containing each of the
            response units found in the correspondences"""
        resp_map = self.response_units = set()
        for response_table in self.stimulus_units.values():
            for response in response_table:
                resp_map.add(response)
    
    def column_heads(self):
        """Returns list of column headers for output"""
        heads = self.trials['column_heads'][:]
        heads.extend(['distance', 'align'])
        return heads
        
    def parse(self, string, table):
        """Given a spelling or pronunciation, lists all the ways it can
        be chunkified into functional units. The functional units are
        itemized as keys in the supplied table. E.g. string 'bred' could
        return [['b', 'r', 'e', 'd'], ['b', 'r', 'ed'], ['b', 're', 'd']]
        because all the individual
        letters are functional units, and in addition 're' and 'ed' both
        may spell single phonemes in some words."""
        whole_length = len(string)
        if not whole_length: return [[]]
        results = []
        #accept_identity = self.correspondences['accept-identity']
        for head_len in range(1, whole_length + 1):
            head = string[:head_len]
            #if head in table  or  accept_identity and head_len == 1:
            if head in table  or  head_len == 1:
                tail_parses = self.parse(string[head_len:], table)
                for tail in tail_parses:
                    result = tail[:]
                    result.insert(0, head)
                    results.append(result)
        return results
    
    def score_parses_sequential(self, stimulus_parse, response_parse):
        """Score spelling using Levenshtein"""
        lev = self.levenshtein
        lev.set_pair(stimulus_parse, response_parse)
        return lev.min_edit_distance()
    
    def score_parses_nonsequential(self, stimulus_parse, response_parse):
        """Score spelling but allow random order and overlap"""
        total = 0
        accept_identity = self.correspondences['accept-identity']
        penalties = self.correspondences['map']
        deletion_penalty = self.correspondences['deletion-penalty']
        align = ''
        for stim in stimulus_parse:
            best_score = None
            best_match = None
            resp_table = penalties.get(
                stim, {stim: 0} if accept_identity else [])
            for resp in resp_table:
                if resp == ''  or  resp in response_parse:
                    if best_score is None or resp_table[resp] < best_score:
                        best_score = resp_table[resp]
                        best_match = resp
            if best_score is None:
                best_score = deletion_penalty
            total += best_score
            align += '[{0}={1}{2}]'.format(
                stim,
                '' if not best_match else best_match,
                '*' if best_score > 0 else '')
        self.last_align = align
        return total
            
    #what happens if stimulus or response is unparsable? should raise error.
    
    def score_parses(self, stimulus_parse, response_parse):
        """Returns score for a single parse of the stimulus against a 
            single parse of the response."""
        if self.scoring['require-order']:
            return self.score_parses_sequential(stimulus_parse, response_parse)
        else:
            return self.score_parses_nonsequential(
                stimulus_parse, response_parse)
    
    def align_last_pair_sequential(self):
        """Returns the alignment for the last pair that was scored via
            Levenshtein"""
        parts = []
        last_cost = 0
        for pr, sp, cost in self.levenshtein.alignment_facts():
            #cumulative, so subtract
            this_cost = cost - last_cost
            sub = pr and sp
            parts.append('[{p}={s}{c}]'.format(
                p=(pr or ''), s=(sp or ''), 
                c=('*' if this_cost else '')))
            last_cost = cost
        return ''.join(parts)
    
    def align_last_pair(self):
        """Returns the alignment for the last pair that was scored."""
        if self.scoring['require-order']:
            return self.align_last_pair_sequential()
        else:
            return self.last_align
    
    def all_parses(self, string, table):
        """Returns all possible parses of string into correspondence units.
            If string has several alternatives separated by | each
            alternative is processed separately, and the union of their
            parses is returned."""
        options = string.split('|')
        all_parses = []
        for option in options:
            option = option.strip()
            option_parses = self.parse(option, table)
            if not option_parses and len(option):
                raise PontoError(
                    'There are insufficient correspondences for "{0}". Please'
                    ' ensure your Correspondences list is complete, or select'
                    ' the option to permit unspecified Identity'
                    ' correspondences like [x=x]'.format(string))
            all_parses.extend(option_parses)
        return all_parses 
        
    def score_response(self, stimulus, response, include_alignment=False):
        """Score one response against its stimulus, returning
             (distance, align)"""
        stimulus_parses = self.all_parses(stimulus, self.stimulus_units)
        response_parses = self.all_parses(response, self.response_units)
        distance = None
        align = None
        for stimulus_parse in stimulus_parses:
            for response_parse in response_parses:
                distance1 = self.score_parses(
                    stimulus_parse, response_parse)
                if  distance == None  or  distance1 < distance:
                    distance = distance1
                    if include_alignment:
                        align = self.align_last_pair() 
        return (distance, align) if include_alignment else distance
    
    def store_per_subject(self, subject_id, stimulus, response, distance):
        if subject_id not in self.subjects:
            self.subjects[subject_id] = {
                'id': subject_id,
                'stimuli': [], 'responses': [], 
                'distances': [], 'all-distances': {}}
        subject = self.subjects[subject_id]
        subject['stimuli'].append(stimulus)
        subject['responses'].append(response)
        subject['distances'].append(distance)
        subject['all-distances'][stimulus] = {response: distance}
    
    def score_row(self, row):
        """Score one row of input. Return copy, with scores appended
            (distance score, then alignment)"""
        row = row[:]
        row_dict = dict(zip(self.trials['column_heads'], row))
        stimulus = row_dict.get('stimulus', None)
        response = row_dict.get('response', None)
        if (not stimulus) or stimulus == '---' or response == '---':
            row.extend(['---', '---'])
            return row
        distance, align = self.score_response(
            stimulus, response, include_alignment=True)
        row.extend([distance, align])
        self.store_per_subject(
            row_dict.get('subject', 'ALL'), stimulus, response, distance)
        return row        
    
    def rows(self):
        """Generator yields a row of results: input row with scores appended"""
        for row in self.trials['rows']:
            yield self.score_row(row)
    
    def summary_column_heads(self):
        """Returns suggested names for columns of a summary table,
            which contains one summary row per subject"""
        columns = ['subject', 'distSum', 'nTrials', 'distMean']
        if self.scoring['do-monte']:
            columns.extend(['distRand', 'improve', 'pRand'])
        return columns
    
    def measure_all_pairings(self, subject):
        """Fills out dict 'all-distances' with distance 
            between each stimulus and response attempted by subject"""
        distances = subject['all-distances']
        for stim in subject['stimuli']:
            for resp in subject['responses']:
                if resp not in distances[stim]:
                    distances[stim][resp] = self.score_response(stim, resp)
    
    def compute_sum_of_distances(self, subject):
        """Returns total error score for this random arrangement of
            responses."""
        sod = 0.0
        distances = subject['all-distances']
        for stimulus, response in zip(
                subject['stimuli'], subject['responses']):
            sod += distances[stimulus][response]
        return sod
    
    def do_monte(self, subject):
        """Use Monte Carlo test to see if subject's spelling is significantly
        better than the score he would get if responses are randomly
        associated with stimuli."""
        self.measure_all_pairings(subject)
        base_distance = subject['dist_sum']
        n_tail = 1
        sum_rearranged = 0  
        n_arrangements = 1
        responses = subject['responses']
        for arr in range(1, 1000):
            random.shuffle(responses)
            rearranged_distance = self.compute_sum_of_distances(subject)
            if rearranged_distance <= base_distance: n_tail += 1
            sum_rearranged += rearranged_distance 
            n_arrangements += 1
        subject['all-distances'] = None
        subject['n-tail'] = n_tail
        subject['dist-rand'] = dist_rand = \
            sum_rearranged / (n_arrangements - 1)
        subject['improve'] = (dist_rand - base_distance) / dist_rand \
            if dist_rand != 0 else '∞'
        subject['p-rand'] = n_tail / n_arrangements
    
    def summarize_subject(self, subject):
        """Summarize findings for this subject, return a list of facts"""
        row = [subject['id']]
        distances = subject['distances']
        distSum = subject['dist_sum'] = sum(distances)
        row.append(distSum)
        nTrials = len(distances)
        row.append(nTrials)
        distMean = distSum / nTrials if nTrials else None
        row.append(distMean)
        if self.scoring['do-monte']:
            self.do_monte(subject)
            row.append(subject['dist-rand'])
            row.append(subject['improve'])
            row.append(subject['p-rand'])
        return row
    
    def summary_rows(self):
        """Generator yields a row of summaries: one subject"""
        for subject in sorted(self.subjects):
            yield self.summarize_subject(self.subjects[subject])
